from geopandas import GeoDataFrame, sjoin

class Base():
    ''' Base class for masking methods '''
    def __init__(self,
                sensitive_gdf,
                population_gdf='',
                population_column='pop',
                container_gdf='',
                max_tries=1000,
                address_points_gdf=''):
        self.sensitive = sensitive_gdf.copy()
        self.crs = self.sensitive.crs
        self._load_population(population_gdf, population_column)
        self._load_container(container_gdf)
        self._load_addresses(address_points_gdf)
        self.max_tries = max_tries


    def _load_population(self, population_gdf='', population_column='pop'):
        '''Loads a geodataframe of population data for donut masking 
        and/or k-anonymity estimation.'''
        if isinstance(population_gdf, GeoDataFrame):
            assert population_gdf.crs == self.crs, "Population CRS does "\
                "not match points CRS"
            self.population = self._crop_gdf(
                population_gdf, self.sensitive)
            self.pop_column = population_column
            self.population = self.population.loc[:,['geometry', self.pop_column]]
            return True
        else:
            self.population = ''
            return False


    def _load_container(self, container_gdf):
        '''Loads a geodataframe of polygons to contain points while donut masking'''
        if isinstance(container_gdf, GeoDataFrame):
            assert container_gdf.crs == self.crs, "Container CRS does "\
                "not match points CRS"
            self.container = self._crop_gdf(
                container_gdf, self.sensitive)
            self.container = self.container.loc[:,['geometry']]
            self.container_filtered = self.container.copy()
            return True
        else:
            self.container = ''
            return False


    def _load_addresses(self, address_points_gdf):
        '''Loads geodataframe containing address data for k-anonymity calculation'''
        if isinstance(address_points_gdf, GeoDataFrame):
            assert address_points_gdf.crs == self.crs, "Address points "\
                "CRS does not match points CRS"
            self.addresses = self._crop_gdf(
                address_points_gdf, self.sensitive)
            self.addresses = self.addresses.loc[:,['geometry']]
            return True
        else:
            self.addresses = ''
            return False


    def _crop_gdf(self, target_gdf, reference_gdf):
        '''Uses spatial index to reduce an input (target) geodataframe to only that which
        intersects with a reference geodataframe'''
        bb = reference_gdf.total_bounds
        x = ((bb[2] - bb[0]) / 10)
        y = ((bb[3] - bb[1]) / 10)
        bb[0] = (bb[0] - x)
        bb[1] = (bb[1] - y)
        bb[2] = (bb[2] + x)
        bb[3] = (bb[3] + y)
        target_gdf = target_gdf.cx[bb[0]:bb[2], bb[1]:bb[3]]
        return target_gdf


    def displacement_distance(self):
        '''Calculate dispalcement distance for each point after masking.'''
        assert isinstance(self.masked, GeoDataFrame), "Data has not yet been masked"
        for index, row in self.masked.iterrows():
            old_coords = self.sensitive.at[index,'geometry']
            distance = row.geometry.distance(old_coords)
            self.masked.at[index,'distance'] = distance
        return self.masked


    def k_anonymity_estimate(self, population_gdf='', population_column='pop'):
        '''Estimates k-anoynmity based on population data.'''
        if not isinstance(self.population, GeoDataFrame):
            self._load_population(population_gdf, population_column)

        assert isinstance(self.sensitive, GeoDataFrame), "Sensitive points geodataframe is missing"
        assert isinstance(self.masked, GeoDataFrame), "Data has not yet been masked"
        assert isinstance(self.population, GeoDataFrame), "Population geodataframe is missing"

        self.population['pop_area'] = self.population.area
        
        if 'distance' not in self.masked.columns:
            self.displacement_distance()

        masked_temp = self.masked.copy()

        masked_temp['geometry'] = masked_temp.apply(
            lambda x: x.geometry.buffer(x['distance']), axis=1)

        masked_temp = self._disaggregate_population(masked_temp)

        for i in range(len(self.masked.index)):
            self.masked.at[i,'k_est'] = int(
                masked_temp.loc[masked_temp['index_2'] == i, 'pop_adjusted'].sum() - 1)
        
        return self.masked


    def k_anonymity_actual(self, address_points_gdf=''):
        '''Calculates k-anonymity based on the number of addresses closer 
        to the masked point than sensitive point'''
        if not isinstance(self.addresses, GeoDataFrame):
            self._load_addresses(address_points_gdf)
            
        assert isinstance(self.sensitive, GeoDataFrame), "Sensitive points geodataframe is missing"
        assert isinstance(self.masked, GeoDataFrame), "Data has not yet been masked"
        assert isinstance(self.addresses, GeoDataFrame), "Address points geodataframe is missing"

        if isinstance(self.addresses, GeoDataFrame) is False:
            raise Exception("Error: missing address point geodataframe.")

        if 'distance' not in self.masked.columns:
            self.displacement_distance()
        
        masked_temp = self.masked.copy()

        masked_temp['geometry'] = masked_temp.apply(
            lambda x: x.geometry.buffer(x['distance']), axis=1)

        join = sjoin(self.addresses, masked_temp, how='left')

        for i in range(len(self.masked)):
            subset = join.loc[join['index_right'] == i,:]
            self.masked.at[i,'k_actual'] = len(subset)

        return self.masked


    def _disaggregate_population(self, target_gdf):
        '''Used for estimating k-anonymity. Disaggregates population within
        buffers based on population polygon data'''
        target = target_gdf.copy()
        target = sjoin(
            target, 
            self.population, 
            how='left')

        target['index_2'] = target.index

        target.index = range(len(target.index))

        target['geometry'] = target.apply(
            lambda x: x['geometry'].intersection(
                self.population.at[x['index_right'],'geometry']), 
                axis=1)

        target['intersected_area'] = target['geometry'].area

        for i in range(len(target_gdf.index)):
            
            polygon_fragments = target.loc[target['index_2'] == i, :]
            
            for index, row in polygon_fragments.iterrows():
                area_pct = row['intersected_area'] / row['pop_area']
                target.at[index,'pop_adjusted'] = row[self.pop_column] * area_pct
        
        return target


    def _containment(self, uncontained):
        '''If a container geodataframe is loaded, checks whether or not masked 
        points are within the same containment polygon as their original locations.'''
        if 'index_right' not in self.sensitive.columns:
            self.sensitive = sjoin(self.sensitive, self.container, how='left')
            self.tries = 0

        self.container_filtered = self._crop_gdf(self.container_filtered, uncontained)

        uncontained = sjoin(uncontained, self.container_filtered, how='left')

        for index, row in uncontained.iterrows():
            if row['index_right'] == self.sensitive.iat[index, -1]:
                self.masked.at[index,'contain'] = 1

        self.tries +=1
        
        if self.tries > self.max_tries:
            for index, row in uncontained.iterrows():
                self.masked.loc[index,'contain'] = 999
            
            print(str(len(uncontained)) + " points were masked but could not be" \
                "contained. Uncontained points are listed as 999 in the 'contain' field")
        
        return True


from osmnx import graph_from_xml, graph_from_bbox, graph_to_gdfs
from osmnx import save_graph_xml
from osmnx.distance import get_nearest_node
from osmnx.utils_graph import  add_edge_lengths, remove_isolated_nodes
from geopandas import GeoDataFrame
from pandas import concat
from networkx import single_source_dijkstra_path_length
from random import choice
from .mask import Base
from multiprocessing import Pool, cpu_count
from hashlib import md5
from copy import deepcopy
from os.path import exists
from numpy import array_split


class Street(Base):
    def __init__(
                self, 
                sensitive_gdf, 
                depth=20,
                extent_expansion_distance=2000,
                max_street_length=500,
                population_gdf='', 
                population_column='pop',
                address_points_gdf='',
                ):
        
        super().__init__(
            sensitive_gdf = sensitive_gdf, 
            population_gdf = population_gdf, 
            population_column = population_column,
            address_points_gdf = address_points_gdf)

        self.buffer_dist = extent_expansion_distance
        self.max_street_length = max_street_length
        self.depth = depth


    def _get_osm(self):
        selection = self.masked.buffer(self.buffer_dist)
        selection = selection.to_crs(epsg=4326)
        bb = selection.total_bounds

        '''
        graph_filename = "osm_" + str(md5(bb).hexdigest()[0:9]) + ".xml"
    
        if exists("data/" + graph_filename):
            self.graph = graph_from_xml("data/" + graph_filename)
        else:
        '''
        self.graph = graph_from_bbox(
            north=bb[3], south=bb[1], east=bb[2], west=bb[0],
            network_type="drive", truncate_by_edge=True)
        #save_graph_xml(self.graph, filepath=graph_filename)
    
        self.graph = remove_isolated_nodes(self.graph)
        self.graph = add_edge_lengths(self.graph)
        self.graph_gdf = graph_to_gdfs(self.graph)

        return True


    def _nearest_node(self, graph_temporary, geometry):
        neighbor_count = 0

        while neighbor_count < 1:
            node = get_nearest_node(
                graph_temporary, (geometry.centroid.y, geometry.centroid.x))
            
            neighbor_count = len(self._find_neighbors(node))
        
            if neighbor_count == 0:
                graph_temporary.remove_node(node)
        
        return node


    def _find_neighbors(self, node):
        neighbors = list(self.graph.neighbors(node))
    
        for neighbor in neighbors:
            length = self.graph.get_edge_data(node, neighbor)[0]['length']
    
            if length > self.max_street_length:
                neighbors = [n for n in neighbors if n != neighbor]

        return neighbors


    def _street_mask(self, node):
        node_count = 0
        distance = 250
        threshold = self.depth
        
        while node_count < threshold:
            paths = single_source_dijkstra_path_length(self.graph, node, distance, 'length')
            node_count = len(paths)
            distance += 250

        nodes = []
        distances =  []
        total_distance = 0
        i = 0
        
        for key, value in paths.items():
            nodes.append(key)
            distances.append(value)
            total_distance += value
            i +=1
            if i == threshold:
                break

        candidate_distance = total_distance / threshold
        
        idx = distances.index(min(distances, key=lambda x:abs(x-candidate_distance)))

        node = nodes[idx]

        return node


    def _apply_street_mask(self, masked_gdf):

        graph_temporary = self.graph.copy()

        masked_gdf['node'] = masked_gdf.apply(
            lambda x: self._nearest_node(graph_temporary, x['geometry']), axis=1)

        masked_gdf['node_new'] = masked_gdf.apply(
            lambda x: self._street_mask(x['node']), axis=1)

        masked_gdf['geometry'] = masked_gdf.apply(
            lambda x: self.graph_gdf[0].at[x['node_new'], 'geometry'], axis=1)
        
        masked_gdf = masked_gdf.drop(['node', 'node_new'], axis=1)
    
        return masked_gdf


    def execute(self, parallel=False):
        self.masked = self.sensitive.copy()

        self._get_osm()
    
        self.masked = self.masked.to_crs(epsg=4326)
        
        if parallel == True:
            self.masked = self.execute_parallel()
        else:
            self.masked = self._apply_street_mask(self.masked)

        self.masked = self.masked.to_crs(self.crs)

        return self.masked


    def execute_parallel(self):
        cpus = cpu_count()
        pool = Pool(processes=8)
        chunks = array_split(self.masked, cpus)

        processes = [pool.apply_async(
            self._apply_street_mask,
            args=(chunk,)) 
            for chunk in chunks]

        masked_chunks = [chunk.get() for chunk in processes]

        return GeoDataFrame(
            concat(masked_chunks), crs={'init': 'epsg:4326'}) 


from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="maskmypy",
    version="0.0.3",
    author="David Swanlund",
    author_email="david.swanlund@gmail.com",
    description="Geographic masking tools for spatial data anonymization",
    long_description=long_description,
    long_description_content_type="text/markdown",
    #url="https://github.com/pypa/sampleproject",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: GIS",
    ],
    python_requires='>=3.6',)


from geopandas import GeoDataFrame, sjoin
from random import random, gauss, uniform
from .mask import Base
from shapely.affinity import translate
from math import sqrt

class Donut(Base):
    """
    Summary line for donut

    Extended description of function.

    Parameters
    ----------
    sensitive_gdf : GeoDataFrame
        A GeoDataFrame of the sensitive points that are to be masked.

    Returns
    -------
    Object 
        An object for donut geomasking

    """
    def __init__(
                self, 
                sensitive_gdf, 
                population_gdf='', 
                population_column='pop',
                max_distance=250,
                donut_ratio=0.1,
                distribution='uniform',
                container_gdf='',
                address_points_gdf='',
                max_tries=1000):
        
        super().__init__(
            sensitive_gdf = sensitive_gdf, 
            population_gdf = population_gdf, 
            population_column = population_column,
            container_gdf = container_gdf,
            max_tries = max_tries,
            address_points_gdf = address_points_gdf)

        self.max = max_distance
        self.distribution = distribution
        self.donut_ratio = donut_ratio
        

    def _random_xy(self, min, max):
        if self.distribution == 'uniform':
            hypotenuse = uniform(min, max)
            x = uniform(0,hypotenuse)

        elif self.distribution == 'gaussian':
            mean = (((max - min) / 2) + min)
            sigma = (((max - min) / 2) / 2.5)
            hypotenuse = gauss(mean, sigma)
            x = uniform(0, hypotenuse)

        elif self.distribution == 'areal':
            hypotenuse = 0
            while hypotenuse == 0:
                r1 = uniform(min, max)
                r2 = uniform(min, max)
                if r1 > r2:
                    hypotenuse = r1
            x = uniform(0, hypotenuse)

        else:
            raise Exception("Unknown distribution") 

        y = sqrt(hypotenuse**2 - x**2)
        
        direction = random()

        if direction < 0.25:
            x = x * -1

        elif direction < 0.5:    
            y = y * -1

        elif direction < 0.75:
            x = x * -1
            y = y * -1

        elif direction < 1:
            pass

        return (x, y)
    

    def _find_radii(self):
        self.masked.loc[:,'radius_min'] = self.max * self.donut_ratio
        self.masked.loc[:,'radius_max'] = self.max


    def _mask_within_container(self): 
        self.masked.loc[:,'contain'] = 0  
        
        while min(self.masked['contain']) == 0:
            
            uncontained = self.masked.loc[self.masked['contain'] == 0, :]
            
            for index, row in uncontained.iterrows():
                x,y = self._random_xy(row['radius_min'], row['radius_max'])
                
                self.masked.at[index, 'geometry'] = translate(
                    row['geometry'], xoff=x, yoff=y)
            
            self._containment(uncontained)

        return True


    def execute(self):
        self.masked = self.sensitive.copy()

        self._find_radii()

        self.masked['offset'] = self.masked.apply(
            lambda x: self._random_xy(
                x['radius_min'], x['radius_max']), axis=1)

        self.masked['geometry'] = self.masked.apply(
            lambda x: translate(
                x['geometry'], xoff=x['offset'][0], yoff=x['offset'][1]), axis=1)

        if isinstance(self.container, GeoDataFrame):
            self._mask_within_container()

        self.masked = self.masked.drop(['offset'], axis=1)
        
        return self.masked



class Donut_MaxK(Donut):

    def __init__(
                self, 
                sensitive_gdf, 
                population_gdf='', 
                population_column='pop',
                max_k_anonymity=0,
                donut_ratio=0.1,
                distribution='uniform',
                container_gdf='',
                address_points_gdf='',
                max_tries=1000):
        
        super().__init__(
            sensitive_gdf = sensitive_gdf, 
            population_gdf = population_gdf, 
            population_column = population_column,
            container_gdf = container_gdf,
            max_tries = max_tries,
            address_points_gdf = address_points_gdf,
            donut_ratio = donut_ratio,
            distribution = distribution)

        self.target_k = max_k_anonymity


    def _find_radii(self): 

        self.population['pop_area'] = self.population.area

        join = sjoin(self.masked, self.population, how='left')
        
        join['max_area'] = join.apply(
            lambda x: self.target_k * x['pop_area'] / x[self.pop_column], axis=1)

        join['min_area'] = join.apply(
            lambda x: (self.target_k * self.donut_ratio) * x['pop_area'] / 
                x[self.pop_column], axis=1)

        join['max_radius'] = join.apply(
            lambda x: sqrt(x['max_area'] / 3.141592654), axis=1)

        join['min_radius'] = join.apply(
            lambda x: sqrt(x['min_area'] / 3.141592654), axis=1)
        
        self.masked['radius_min'] = join.apply(
            lambda x: x['min_radius'], axis=1)
        self.masked['radius_max'] = join.apply(
            lambda x: x['max_radius'], axis=1)



class Donut_Multiply(Donut):

    def __init__(
                self, 
                sensitive_gdf, 
                max_distance=250,
                population_gdf='', 
                population_column='pop',
                population_multiplier = 0,
                donut_ratio=0.1,
                distribution='uniform',
                container_gdf='',
                address_points_gdf='',
                max_tries=1000):
        
        super().__init__(
            sensitive_gdf = sensitive_gdf, 
            max_distance = max_distance,
            population_gdf = population_gdf, 
            population_column = population_column,
            container_gdf = container_gdf,
            max_tries = max_tries,
            address_points_gdf = address_points_gdf,
            donut_ratio = donut_ratio,
            distribution = distribution,
                )

        self.pop_multiplier = population_multiplier - 1
        

    def _find_radii(self):
        self.population['pop_area'] = self.population.area

        join = sjoin(self.masked, self.population, how='left')
        
        pop_min = min(join[self.pop_column])
        pop_max = max(join[self.pop_column])
        pop_range =  pop_max - pop_min
        
        join['pop_score'] = join.apply(
            lambda x: (1 - (x[self.pop_column] - pop_min) / pop_range) \
                * self.pop_multiplier, axis=1)

        self.masked['radius_max'] = join.apply(
            lambda x: (x['pop_score'] * self.max) + self.max, axis=1)

        self.masked['radius_min'] = self.masked.apply(
            lambda x: x['radius_max'] * self.donut_ratio, axis=1)
